/* Copyright IBM 1995,2003 */
/* $Id: site.mcr.proto,v 1.695.2.19 2010/11/05 18:26:13 mcnabb Exp $ */

/******************************************************************************/
/* Platform and Architecture Defines:                                         */
/*                                                                            */
/* Define the platform (operating system) and architecture (instruction set)  */
/* settings. Only one platform and one architecture may be defined. Not all   */
/* combinations are supported.                                                */
/*                                                                            */
/* Supported platform settings:                                               */
/*                                                                            */
/*    GPFS_AIX                                                                */
/*    GPFS_LINUX                                                              */
/*    GPFS_SOLARIS                                                            */
/*    GPFS_WINDOWS (+)                                                        */
/*                                                                            */
/* Supported architecture settings:                                           */
/*                                                                            */
/*    GPFS_ARCH_I386                                                          */
/*    GPFS_ARCH_IA64                                                          */
/*    GPFS_ARCH_POWER                                                         */
/*    GPFS_ARCH_PPC64                                                         */
/*    GPFS_ARCH_X86_64                                                        */
/*    GPFS_ARCH_SPARC_V9                                                      */
/*    GPFS_ARCH_S390X                                                         */
/*                                                                            */
/* + GPFS_WINDOWS uses this file for COMMON_DEFINES, COMMON_TEMPSHIP_DEFINES, */
/*   and COMMON_NEVERSHIP_DEFINES. Windows platform and architecture defines  */
/*   are specified in ./ts/windows/site_mcr.h.                                */
/******************************************************************************/


/* use LINUXPR to turn on SCSI-3 Persistent Reserve support */
#define LINUXPR

/* use PR_4_IPV6 to turn on PR support for IPV6 environment */
/* #define PR_4_IPV6 */

/* use PERCPU_BUFFER_TRACE_ONAIX to turn on per cpu buffer tracing */
#define PERCPU_BUFFER_TRACE_ONAIX

#ifdef GPFS_SOLARIS
/*****************************************************/
/* Solaris specific config defines begin.            */
/*****************************************************/

/******************************************************************************/
/* Solaris OS version defines:                                                */
/*                                                                            */
/* Define the Solaris target OS version.  Some things change in the kernel    */
/* with subsequent releases of the OS.                                        */
/* Supported target OS settings:                                              */
/*                                                                            */
/*    GPFS_SOL_09                                                             */
/*                                                                            */
/******************************************************************************/
#define GPFS_SOL_09

/*****************************************************/
/* Solaris OS version defines end.                   */
/*****************************************************/
#undef linux
#undef GPFS_LINUX
#undef GPFS_ARCH_I386
#endif  /* GPFS_SOLARIS */

#ifdef GPFS_LINUX
/*****************************************************/
/* Linux portability layer defines begin.            */
/* Customer editable portion of config file.         */
/*****************************************************/

/* use #define SRC_BUILD to build the open source kernel module */
/* use #define SRC_BUILD and LIBSRC_BUILD to build the open source libgpfs */
#define SRC_BUILD_PLACEHOLDER
#define LIB_BUILD_PLACEHOLDER

/* use GPL_KBUILD to turn on the GPL layer Kbuild mechanism */
#define GPL_KBUILD
/* use LXTRACE_RELAY to turn on relayfs trace */
#ifndef GPFS_ARCH_POWER
#define LXTRACE_WANT_RELAY
#endif

/* use SHIP_SOBAR to turn on ship of Scale Out Backup and Restore components */
#define SHIP_SOBAR

/******************************************************************************/
/* Linux Distribution Defines:                                                */
/*                                                                            */
/* Define the Linux distribution setting. Only one Linux distribution may be  */
/* defined.                                                                   */
/*                                                                            */
/* Supported Distribution settings:                                           */
/*                                                                            */
/*    REDHAT_AS_LINUX                                                         */
/*    SUSE_LINUX                                                              */
/*                                                                            */
/* Select REDHAT_AS_LINUX for RedHat Enterprise Linux (AS|ES|WS)              */
/* Select SUSE_LINUX for SuSE Enterprise Server                               */
/*                                                                            */
/* Unsupported, use on your own risk:                                         */
/*    KERNEL_ORG_LINUX                                                        */
/*                                                                            */
/******************************************************************************/

/* Linux Kernel Version Defines:                                              */
/*                                                                            */
/* Define the Linux kernel version and kernel includes directory settings.    */
/* Only one Linux kernel version and one kernel includes directory may be     */
/* defined.                                                                   */
/*                                                                            */
/* The Linux kernel version is determined from the first four components of   */
/* the kernel string. It is important to note that the second, third and      */
/* fourth components are specified with two digits. Any component less than   */
/* 10 must be preceded by a '0', while any component greater than or equal to */
/* 100 should specify '99'.                                                   */
/*                                                                            */
/* Examples:                                                                  */
/*                                                                            */
/*    2.6.5-7.139 would be 2060507                                            */
/*    2.6.16.20   would be 2061620                                            */
/*                                                                            */
/* The kernel includes directory is where GPFS will find the kernel headers   */
/* for the kernel specified by the LINUX_KERNEL_VERSION define. The kernel    */
/* includes location varies by Linux distribution.                            */
/*                                                                            */
/* Examples:                                                                  */
/*                                                                            */
/*    /lib/modules/`uname -r`/build/include                                   */
/*    /lib/modules/`uname -r`/source/include                                  */
/*    /usr/src/linux/include                                                  */
/*                                                                            */
/******************************************************************************/
#ifndef GPFS_ARCH_POWER
KERNEL_HEADER_DIR = /lib/modules/`uname -r`/build/include
#endif

/******************************************************************************/
/* Linux Build Directory                                                      */
/*                                                                            */
/* If you are running a vanilla kernel, where in the headers and the build    */
/* directories are in the standard place (i.e. under /lib/modules) then       */
/* running "make Autoconfig" will set the correct values for                  */
/* KERNEL_BUILD_DIR variable.                                                 */
/*                                                                            */
/* The kernel build directory is where the kernel build tools reside.         */
/* These tools are required to build the GPFS modules for Linux.              */
/* Examples:                                                                  */
/*    /lib/modules/`uname -r`/build                                           */
/*                                                                            */
/******************************************************************************/

/* Flags for building with desired Linux kernel patches.
 * These are patches for specific problems we've discovered
 * with the linux kernel and are providing interim patches
 * for.  Each patch is described below and the desired patches 
 * should be uncommented under LINUX_PATCH_DEFINES. 
 *
 * -DPREEMPT_RT_LINUX_PATCH  For Ingo Molnar's PREEMPT_RT patch
 *
 * Descriptions and downloads for these patches may be found at 
 * http://sourceforge.net/tracker/index.php?func=detail&aid=1153778&group_id=130828&atid=719124
 */
LINUX_PATCH_DEFINES = \
/*  -DPREEMPT_RT_LINUX_PATCH */

/*****************************************************/
/* Linux portability layer defines end.              */
/* Everything below here is irrelevent to customer   */
/* compiles of the portability layer.                */
/*****************************************************/
#endif /* GPFS_LINUX */

/* Descriptions of variables used by the build process:

   The following variables are lists available to all makefiles:
     DESTDIR              Root of the ts directory in the build tree.
     CDEBUGFLAGS          Passed to all compiles of code that goes in mmfsd
                            or other process-level code.  Empty by default;
                            intended to be used temporarily within the
                            Imakefile in a directory to turn on generation of
                            assembly listings, say.
     KCDEBUGFLAGS         Passed to all compiles of code that goes in kernel
                            extensions.  Empty by default.
     OPTFLAGS             Compiler optimization flags for daemon code and
                            code that is compiled once and bound into both
                            the daemon and the kernel.  Typical values are -g
                            or -O.
     KOPTFLAGS            Compiler optimization flags for code used exclusively
                            in kernel modules.
     ARCH_DEFINES         -D defines describing the machine architecture
     DEFAULT_DEFINES      List of -D defines used for shipped product code.
                            The complete list of possible -D values is given
                            below.
     TEMPSHIP_DEFINES     List of -D defines used for debuging, not shipped
                            in product code after first few PTFs of new release.
     NEVERSHIP_DEFINES    List of -D defines used for testing, never shipped
                            in product code.
     SITE_DEFINES         List of the complete set of -D defines used by each
                            compilation in the build.  For testing, this will
                            be the concatenation of DEFAULT_DEFINES,
                            ARCH_DEFINES, TEMPSHIP_DEFINES, and
                            NEVERSHIP_DEFINES.  For the shipped product,
                            this will just contain DEFAULT_DEFINES and
                            ARCH_DEFINES (and TEMPSHIP_DEFINES for a
                            little while).

     KERNEL_HEADER_DIR    Directory where Linux kernel header files are kept

     CFLAGS32             Special compiler flags used for building 32-bit
                          code on a 64-bit platform when API_32BIT is defined
     SHLDFLAGS32          Special linker flags used for linking shared 32-bit
                          libraries on a 64-bit platform when API_32BIT 
                          is defined

   The following variables are either defined or undefined by this file.
   Tests of these are used in Imakefiles to conditionally control which
   lines are processed.  For almost all of these there is a
   corresponding -Dxxx that must be specified in SITE_DEFINES.

     DMAPI                #define this to cause code implementing the DMAPI
                            interface to GPFS to be built.  Requires -DDMAPI
                            in SITE_DEFINES.
     FAR_TRACE            #define this to cause calls to the routine that
                            actually does TRACEn to be placed far away from
                            the tests of whether tracing is enabled.  When
                            tracing is off, this results in fewer I-cache
                            misses, and therefore better performance.  Also
                            controls several compiler options that are needed
                            to get the correct code generated.  Cannot be
                            used with -g.
     LINUX_KERNEL_VERSION  #define this to match the version of the Linux
                            include files being used for the applicable
                            architecture type.  This is always an eight
                            digit decimal number.  The first four
                            digits give the major and minor kernel version
                            numbers.  For Linux 2.4 these will always be
                            0204, with the leading zero omitted.  The next
                            two digits give the kernel subversion.  The last
                            two digits give the patch level.  Thus the 
                            LINUX_KERNEL_VERSION of 
                              linux-2.4.18-14 is 2041814
                            while
                              linux-2.4.19    is 2041900

                            The compile-time symbol LINUX_KERNEL_VERSION 
                            will be set to the same value as the make-time 
                            symbol LINUX_KERNEL_VERSION.  
                            The GPFS_KERNEL_VERSION macro uses the same 
                            encoding of kernel version as does 
                            LINUX_KERNEL_VERSION, and can be used to 
                            conditionally compile version dependent code.
     GPROF_DAEMON         #define this to compile and link the daemon for
                            gprof.  Use 'mmfsadm test gprof on|off' to control
                            when profiling is enabled.  (Not tested in years,
                            may not work.)
     TRY_AND_BUY          #define this to build a "try and buy" version of
                            GPFS with a timeout license key
     MMPMON               #define this to build support for MMPMON remote
                            performance monitoring services.  Requires -DMMPMON
                            in SITE_DEFINES.
     MMPMON_CUMULATIVE    #define this to build support for MMPMON remote
                            performance monitoring cumulative all-time stats.
			    Requires -DMMPMON in SITE_DEFINES.
     MMPMON_HISTOGRAMS    #define this to build support for MMPMON remote
                            performance monitoring latency histogram data.
			    Requires -DMMPMON *and* -DMMPMON_HISTOGRAMS
                            in SITE_DEFINES.
     EXPORTER_EVENTS      #define this to enable the APPEND_EXPORTER_EVENT code
			    originally used by the Java GUI
     SMB_LOCKS            #define this to add SMB share mode and oplock support
                            Requires -DSMB_LOCKS in SITE_DEFINES
     CCL                  #define this to add support for CIFS Compatibility 
                           Layer.  Requires -DCCL in SITE_DEFINES
     KTRACE               #define this to build trace device into Linux
                            kernel and allow trace statements in kernel.
                            Requires -DKTRACE in SITE_DEFINES
     NO_ASM_UTIL          #define this to prevent the use of assembler versions
                            of utility routines in classes/basic directory
     CROSS_COMPILE        #define this when cross-compiling to avoid running
                            tools that must be built for the target system but
                            run on the build system (e.g. chksystypes program)
     NOKREGPARMS          #define this to build gpfs compatable for a
			    custom built 2.6 kernel with CONFIG_REGPARMS
			    turned off.
     VDISK                #define this to enable support for software RAID
                            virtual disks.  Requires -DVDISK in SITE_DEFINES.
     MMANTRAS             #define this to build MMANTRAS monitoring framework
                            and SNMP subagent.  Requires installation of
                            NetSNMP base and development packages.
     MMANTRAS_CIM         #define this to build CIM provider. Must be used with
                            MMANTRAS. Requires installation of OpenPegasus base 
                            package.
     LOCK_TRACING         #define this to enable the tracing tool for mutex, 
                            condvar, and sxlock.
                            Supported only for Linux and Intel machines. The 'mmfsadm
                            test lockTrace' is used to enable this feature.
     API_32BIT            #define this when building 32-bit versions of 
                            certain shared libraries on a 64-bit platform.
                            Currently only implemented on Linux/x86_64.


   Descriptions of all -Dxxx values that control building GPFS features.

   Standard system features:  Add to COMMON_DEFINES
     -DAlternateID        Enable support for specifying an alternate login name
                            to be used by the mm commands on rsh/ssh calls.
     -DAUTOMOUNT_REWORK   Invokes automounter via autofs control script.
     -DAPI_32BIT          Enable support for handling 32-bit syscalls/ioctls
                            when running on a 64-bit kernel.
     -DBLOCK_RESTORE      Add support for block level incremental
	                    backup & restore.
     -DDESCQUORUM_OVERRIDE Add support for manual overrides of the stripe
                             group descriptor quorum.
     -DDIO                Direct I/O support
     -DDIO_XW             Define this to allow direct I/O writes to
                            allocate new data blocks rather than falling
                            back to normal I/O, provided that write request
                            is block-aligned block multiple.  DIO will use
                            xw byte-range lock rather than whole-file sx
                            lock for block-sized I/O.
     -DDISK_LEASE_DMS     Add support for a deadman switch.  When disk leasing
                            is on, a kernel timer will be turned on, to be
                            restarted whenever a new lease is obtained.  If
                            the lease cannot be renewed, and timer expires,
                            and there's still some outstanding I/O requests,
                            the node is paniced.
     -DDMAPI              Activate DMAPI code (also define DMAPI).
     -DDMAPI_MOVERS       Activate DMAPI parallel movers code.
     -DDMAPI_PREFETCH_DIRATTRS Active parallel dm_get_dirattr and dm_get_dmattr
     -DDMM_DISK           Enable support for Device-Mapper Multipath (DMM) disks
     -DEXPORTER_EVENTS    Enable the APPEND_EXPORTER_EVENT code
			    originally used by the Java GUI
     -DFILESETMAPPING_API Introduces the API function to translate
                            Fileset Name => ID
     -DFINE_GRAINED_DIR_LOCKING Use name-based tokens to control sharing
			    when multiple nodes are creating and
			    deleting files in the same directory.
     -DFSCK_INSTRUMENT    Enables debug display of instrumentation of fsck
                            operations.  Controlled by mmfsck option -xz.
     -DGPFS321_ADMIN      Enable support for admin changes that will ship with
                            the Windows PTF.
     -DLC_FENCING         Enable LC fencing of VSD disks
     -DMMAP_DIO           Change mmap page-out to use direct I/O to write
                            the page if a GPFS data buffer is not available.
                            This removes the requirement that read/write
                            mapped pages always have a GPFS data buffer to
                            back them.  Requires -DDIO.
     -DMMPMON             Enable MMPMON performance monitoring
			    (also enable -DMMPMON.)
     -DMMPMON_CUMULATIVE  Enable MMPMON performance monitoring cumulative
			    all time stats (also enable -DMMPMON and
			    define MMPMON)
     -DMMPMON_DAEMON_PORT Enable code that allows mmpmon/EE connections via
                            the standard daemon port.  Authentication done by
                            llcomm.  Compatible with MMPMON_NAMED_SOCKET.
     -DMMPMON_HISTOGRAMS  Enable MMPMON performance monitoring latency
			    histograms (also enable -DMMPMON and
			    define MMPMON + MMPMON_HISTOGRAMS.)
     -DMMPMON_NAMED_SOCKET Enable MMPMON named socket instead of the
                            independent EE network port.  Mutually exclusive.
     -DMMPMON_RPC         Enable MMPMON rpc interface
     -DNFS4_ACL           Add support for NFSv4 ACLs.
     -DNFS4_VCM           Add support for the AIX VCM (NFSv4 locking)
     -DPDEFRAG            Add support for parallel defrag
     -DPERF_STATS         Additional performance statistics for benchmark runs
     -DPOLICY_RESTORE     Initial support for new RESTORE policy attributes
     -DPOW2_COND_SHIFT    Enable code that implements multiplies and divides
                            by shifts when possible.
     -DSANERGY            Enable SANergy interfaces.
     -DSFS_PERF_SKIP_SYNC Performance improvement to skip multiple sync calls	
                            for the same file.
     -DSFS_PERF_USE_SMALL_BUFFER try to use small buffers to avoid reading
                            data that might not be referenced.
     -DSINGLE_PORT        Enable support for use of a single port number.
     -DSMB_LOCKS          Add support for SMB open locks and oplocks.
     -DUIDREMAP           Enable support for UID remapping in multicluster
                            environment.
     -DINODE64_PREP       Add support for full 32 bit inode numbers to provide
                          greater than 2^31 files per filesystem.
     -DKERNEL_SYNC_DESTROY_EVENT Support to destroy sync event from kernel


   Standard machine dependent system features
     -D__FULL_PROTO       Used by AIX header files to ensure that prototypes
                          of functions include full parameter list.  Needed
                          since these functions have a C calling convention and
                          thus must always be defined with their full proto.
     -DSPAIXLVL=432       AIX level. Must use this option when compiling on
                            AIX 4.3.2 (with new compiler).
     -DSPAIXLVL=433       Must use this option when compiling on AIX 4.3.3
     -DSPAIXLVL=516       Must use this option when compiling on AIX 5.1.6+
     -DUSE51BREGS         AIX: Use AIX 5.1.B register semantics in assembler
                            code.  Must be turned on for code that will
                            run on AIX 5.x.
     -DFAR_TRACE          LINUX: Generate calls to the TRACE subroutine
                            out-of-line with respect to the tests of whether
                            tracing is enabled.  Do not explicitly include
                            this in the list of -D options; it will be
                            provided automatically if FAR_TRACE or
                            KFAR_TRACE is #defined.
     -DFAR_CODE           LINUX: Enable the BEGIN_FAR_CODE/END_FAR_CODE macros
                            to place arbitrary uncommonly executed code
                            out-of-line.  Requires -DFAR_TRACE.  Do not
                            explicitly include this in the list of -D
                            options; it will be provided automatically
                            if FAR_TRACE or KFAR_TRACE is #defined.
     -DLIMIT_KSTACKS      Generate code that is less efficient in the GPFS
                            kernel module, but that uses less kernel stack
                            space.  Needed on i386 for Linux kernel versions
                            at or beyond 2.6.6.
     -DCHECK_KXTIMESTAMP  Enable code that a) links both mmfsd and mmfs with
                            an object that contains a timestamp b) checks very
                            early during daemon startup that kernel and
                            userspace code have the same timestamp, i.e. the
                            kernel extension loaded in memory is from the same
                            build as the currently starting daemon.  The
                            intent is to prevent hard-to-diagnose problems
                            arising (on AIX, most often) from having a stale
                            kernel extension loaded after installing a new
                            build.  Will only be enabled in ADE/ODE builds,
                            although CVS builders can enable it as well.
     -DGPFS_FRAME_POINTER Tell GPFS code that -fomit-frame-pointer was not
                            specified, so frame pointer is available. Use
                            either -fomit-frame-pointer or -DGPFS_FRAME_POINTER
                            when defining OPTFLAGS

   Defines for new upcoming features:
     -DPERCPU_BUFFER_TRACE_ONAIX	Per CPU buffer tracing, start from AIX
     -DLINUXPR            Persistent Reserve support on Linux
     -DAIXPR              Persistent Reserve support on AIX
     -DPR_4_IPV6          Persistent Reserve support for IPV6 environment
     -DCCL                Add support for CIFS Compatibility Layer.
     -DEE_WATCHDOG_START  Start the EE watchdog thread (checks for anomolies)
                            This feature is used by mmantras.
     -DNFS4_CLUSTER       Add support for NFS Version 4.
                            Requires the 2.4.18 nfsv4 kernel.
     -DP_NFS4             Add support for (p)NFS Version 4.1
                            Requires -DNFS4_CLUSTER to be on and
                            Linux 2.6.26 kernel with CONFIG_NFSD_V4_1.
     -DVDISK              Enable support for software RAID virtual disks.
                            Requires #define VDISK.  Only supported on
                            Linux and AIX.
     -DVDISK_LOOP_DEVICES Enable adjustment of Linux loop devices to make them
                          suitable for use as fake pdisks.  Experimental and
                          development use only!  Requires -DVDISK 
     -DOPENSSL_ENGINES    Enable support for OpenSSL hardware acceleration.
     -DMUST_ALIGN         For architecture that requires alignment, specially
                          when handling RPC replies from OS with different
                          endian byte ordering.
     -DSIDMAPPING         Enable code for mapping between Windows SIDs and Unix
                          user and group IDs.  This code is is driven primary 
                          by Windows VFS code, but the define has to be enabled
                          on all platforms (for opaque attr backup/restore).
     -DDEVOPEN_REWRITE    Rewrite of devOpen routine.
     -DGPFS_CACHE         Enables support for using GPFS as a caching filesystem
                          for a remote cluster. Uses (p)NFS as the protocol for 
                          data movement into (and out of) the cache. Supports
                          asynchronous updates and disconnected operations.
                            Requires #define GPFS_CACHE
     -DENABLE_IMMUTABLE_FILES Enable code for immutable files & retention
     -DDMAPI_SYNC_DESTROY_EVENT   DMAPI sync destroy event
     -DUSER_EXIT_ENTENSIONS Enable user exit script for event notification
     -DFASTEA             Enabled 64bit inode scan with extended attributes
     -DDISKEA             Enable storing extended attributes in the inodes
     -DDISKEA_OVERFLOW    Enable overflow blocks support for DISKEA
	                    NOTE: DISKEA must also be defined.
     -DREGEA              Enable extended attribute registry

     -DTSC_SCATTERED_RECEIVE_REPLY_BUFFS
                          Enables support to read tscomm RPC data into multiple
                          scattered buffer areas.  Currently used to process
                          GetData replies for VDISK data (data coming from an
                          NSD client to an NSD server, headed for disk.)  
                          Used to eliminate reading the data into a contiguous
                          buffer merely to copy that data into the scattered
                          VDISK Vio buffers.
     -DTSC_SCATTERED_SEND_REPLY_BUFFS
                          Enables support to write tscomm reply data from multiple
                          scattered buffer areas.  Currently used to process
                          replies to nsdMsgRead for VDISK data (data coming to an
                          NSD client from an NSD server.)  
                          Used to eliminate copying the data from the scattered
                          VDISK Vio buffers into a contiguous buffer for transmission.
     -DINODE64_FULL       Add support for 48 bit inode numbers with 16 bits
                          reserved for flags to provide greater than 2^32-1
                          files per filesystem.
     -DNLINK32            Add support for 32 bit wide nlink value.
     -DGPFS_QDIRTY        Quota performance feature
     -DRESTORE_DMAPI      Add support for reading/writing DM attributes along with other
                          extended attributes using gpfs_igetattrsx()/gpfs_iputattrsx()
     -DUSE_POPEN_CLASS    Use the Popen class as a wrapper for SCN and
                          other fork/exec code. This feature should be enabled on all
                          plaforms in GPFS 3.4 (or later). Windows will enable this 
                          feature in a GPFS 3.3 PTF.
     -DWINDOWS_FORK_FREE  WINDOWS: Enabled code that replaces fork/exec sequences
                          with the Windows CreateProcess API. Windows will enable
                          this feature in a GPFS 3.3 PTF.
                          USE_POPEN_CLASS is a prerequisite for WINDOWS_FORK_FREE.
     -DCTDB               Enabled CTDB in support of SONAS for Linux platforms.


   Defines for non GA versions of the code but turned on for our web
   based "try and buy" version of GPFS.
     -DTRY_AND_BUY        Build a "try and buy" version of 
                            GPFS with a timeout license key


   Defines for Normal testing, only for early shipped systems:
   Add to COMMON_TEMPSHIP_DEFINES
     -DDBGASSERTS         Generate code for DBGASSERT macros
     -DNEW_DELETE_DEBUG   Override global operators new and delete, to assist
                            in finding uninitialized dynamic storage, writing
                            beyond the end of an allocated object, reference to
                            deleted objects, and storage leaks.  The 'mmfsadm
                            test newdebug' command enables some of these
                            features.
     -DTRACK_USECOUNT     Code to track contributing threads to a stripe
                            group's use count.
     -DTRIGGERS           Respond to 'test trigger xxx' commands
     -DVERBOSETRACE       Generate code to produce traces for trace levels > 10

   Defines for Normal testing, never for shipped systems:
   Add to COMMON_NEVERSHIP_DEFINES
     -DENTRYEXIT_TRACE    Enable routine entry/exit tracing that is controlled
                           by mmfsadm trace dentryexit/kentryexit <n>
     -DINSTRUMENT_LOCKS   Enable summary lock instrumentation for mutexes
                            and lock objects.  Requires -DPTHREAD_STATS.
     -DINSTRUMENT_TIME_LOCKS
                          Enable summary lock instrumentation with sleep time
                            statistics.  Requires -DINSTRUMENT_LOCKS.
     -DPTHREAD_STATS      Generate code to count how many times various
                            paths are taken through thread, mutex, and
                            condition variable code.
     -DTRACE_IO_DATA      If RTPERF trace level is >= 3, dump the first few
                            words of data read or written by any disk I/O
                            operation.  This is a security exposure, and should
                            not be turned on in production.
     -DNO_ASM_SYNCH       Disable implementation of fast versions of mutex
                            acquire/release in assembler.  Enables trace
                            calls and some debugging code under -DDBGASSERTS
                            that is not supported by the assembler
                            implementation.  Required if -DTRACK_RESOURCE_USAGE
                            is used.
     -DKSTACK_CHECK       Keep track of how much kernel stack space is used
                            within GPFS kernel modules.  Print tracebacks of
                            the point of maximum stack usage if it exceeds a
                            threshhold.  Can only be used on Linux.
     -DMMDEBUG_MSGS       Enable additional messages in the admin scripts.

   Defines for weird test conditions, not normally used:
     -DINSTRUMENT_DETAIL_LOCKS
                          Also enable per-mutex lock instrumentation.
                            Requires -DINSTRUMENT_LOCKS.
     -DTM_DETAIL_STATS    Extra counters to track which token prefetches
                            are useful and which are not.
     -DDIRCHECKASSERTS    Check validity of directory disk data structures
                            frequently, and assert when an inconsistency is
                            found.  Expensive, so separated from DBGASSERTS.
     -DPOW2_FORCE_SHIFT   Do only shifts for some math. Requires that all
                            GPFS block sizes be power of 2.
                            Since VTS uses 384K blocksize, this is not a
                            shipable option.
     -DTIME_MUTEXES       Time how long mutexes are held.  Summary produced
                            by 'mmfsadm dump lstat'.  
                            Requires -DINSTRUMENT_LOCKS.
     -DGPFS_PRINTF        On Linux, send trace output to the console via printk
                            instead of collecting trace output using ltrace.
                            (requires GPFS_LINUX).
     -DTSCAN              Enable trace points used by the tscan utility to
                            measure CPU usage by system call.
     -DTRACK_RESOURCE_USAGE
                          Keep track of resources (locks, mutexes, buffers,
                            threads, etc.) held by a thread and verify that
                            resources are acquired in a consistent order.
                            Requires -DNO_ASM_SYNCH.
     -DSUPPORT_FOR_DOS_STATFS  Hack for DOS client fsstat
     -DDACSPY             Counts the number of times objects of each of
                            the dynamic array classes are created.
                            The classes are defined via the DECL_ARRAY_CLASS
                            macro in DAC.h.  This optional instrumentation
                            allows for determining which objects being
                            allocated on the heap may require allocation
                            optimization such as via a linked list of
                            preallocated objects.
     -DMALLOC_DEBUG       Generate code to track the file and line number from
                            which each block of MALLOCed storage was allocated
     -DNEW_DELETE_TRACEBACKS
                          Save the call stack on memory allocations tracked
                            by -DNEW_DELETE_DEBUG.  Not supported on Linux.
     -DDIR_VERSION_NUMBERS
                          Increment one of the reserved words in directory
                            block headers each time the block is written to
                            disk.  Helps to sequence directory updates in time
                            across multiple nodes.  Should never be turned on
                            except on file systems that will be reformatted.
     -DTEST_SMALL_DISKS=x When creating a file system or adding disks,
                            artificially limit the disk size to x megabytes.
                            Useful for testing ENOSPC conditions without having
                            to fill up large disks or creating new, smaller
                            logical volumes.
     -DTEST_DELDISK       Ignore all but the first disk listed in mount
                            options.  This is useful when testing so that
                            the tsdeldisk and tsrpldisk commands can be used
                            without the "mm" commands.
     -DNO_VFS_STATS       Do not generate any code to generate statistics for
                            the 'mmfsadm vfsstats {enable|disable|reset|show}'
                            family of commands.
     -DSTOP_TRACE_ON_FAILURE Set all trace flags to 0 as soon as LOGASSERT
                            is triggered or deamon cleanup begins.
     -DDEBUG_SPINLOCK     Define this if Linux kernel is built with
                            CONFIG_DEBUG_SPINLOCK.
     -DFILEIDCMD_PERF     Define this to enable perf measurements for mmfileid
                            command
     -DKTRACE             Define this to build trace device into Linux
                            kernel and allow trace statements in kernel
                            (also define KTRACE).
     -DENABLE_SKIP_DATA_MOVE Enable the 'mmfsadm test disableDataXfer 0/1'
                            command to skip moving data between user buffers
                            and GPFS buffers and to skip data I/O.  Allows
                            simulating a higher metadata load than the actual
                            data I/O configuration would support.
     -DNOW_TRACING        Generate code to trace all calls of now(), the
                            routine that fetches the current time of day.
                            Set TRACE_BASIC to 1 to see the traces, which
                            include the function, file, and line number of
                            each call.
     -DVDISK_PRINTFS      Produce verbose messages using printf in vdisk code.
                            Requires -DVDISK.
     -DSC06               Code required for Supercomputing'06 demo only
     -DLOCK_TRACING       Enable the tracing tool for mutex, condvar, and sxlock.
                            Supported only for Linux and Intel machines. The 'mmfsadm
                            test lockTrace' is used to enable this feature.
     -DPTH_DISABLE_CANCELLATION
	                  Disable pthread cancellation capabilities and disable
			    pthread_testcancel() in Windows SUA only.
     -DWITHDRAW_INODES    Allow removing a range of inodes from service to
                            enable testing large inode numbers without having
                            to allocate them all.
     -DVALIDATE_LISTS     Generate code to check the consistency of the
                            doubly-linked lists provided by the DListHead_t
                            class.  Check lists on entry and exit of all
                            operations that modify such a list.
     -DDISABLE_MB_QUEUE_SPLIT  Unless DISABLE_MB_QUEUE_SPLIT is defined,
                            multiple entries will be created for worker
			    threads in the arrays used for kernel-to-daemon
			    mailboxes.  This can improve performance on
			    systems with many processors.  Define
			    DISABLE_MB_QUEUE_SPLIT to use a single mailbox
			    for testing purposes.  */

CDEBUGFLAGS =
KCDEBUGFLAGS =

#ifdef GPFS_ARCH_SPARC_V9
ARCH_DEFINES = -DGPFS_ARCH_SPARC_V9 \
  -D__64BIT__ \
  -DALIGNED_DATA_ACCESS \
  -DGPFS_BIG_ENDIAN \
  -DAPI_32BIT
#endif

#ifdef GPFS_ARCH_POWER
ARCH_DEFINES = -DGPFS_ARCH_POWER
#endif

#ifdef  GPFS_ARCH_I386
ARCH_DEFINES = -DGPFS_ARCH_I386 -DGPFS_LITTLE_ENDIAN
#endif

#ifdef  GPFS_ARCH_IA64
ARCH_DEFINES = \
  -D__64BIT__ \
  -DALIGNED_DATA_ACCESS \
  -DGPFS_ARCH_IA64 \
  -DGPFS_LITTLE_ENDIAN \
  -DAPI_32BIT
#endif

#ifdef  GPFS_ARCH_PPC64
ARCH_DEFINES = -DGPFS_ARCH_PPC64 -D__64BIT__ -DALIGNED_DATA_ACCESS
#endif

#ifdef  GPFS_ARCH_X86_64
ARCH_DEFINES = -DGPFS_ARCH_X86_64 -D__64BIT__ -DGPFS_LITTLE_ENDIAN \
               -DAPI_32BIT
#endif

#ifdef  GPFS_ARCH_S390X
ARCH_DEFINES = \
  -DGPFS_ARCH_S390X \
  -D__64BIT__ \
  -DALIGNED_DATA_ACCESS \
  -DAPI_32BIT
#endif

#ifdef LINUXPR
  DEFINE_LINUXPR = -DLINUXPR
#else
  DEFINE_LINUXPR =
#endif

#ifdef PR_4_IPV6
  DEFINE_PR_4_IPV6 = -DPR_4_IPV6
#else
  DEFINE_PR_4_IPV6 =
#endif

#ifdef PERCPU_BUFFER_TRACE_ONAIX
  DEFINE_PERCPU_BUFFER_TRACE_ONAIX = -DPERCPU_BUFFER_TRACE_ONAIX
#else
  DEFINE_PERCPU_BUFFER_TRACE_ONAIX =
#endif

/***********************************/
/* Defines common to all platforms */
/***********************************/
COMMON_DEFINES = \
  -DAIXPR \
  -DAlternateID \
  -DAUTOMOUNT_REWORK \
  -DBLOCK_RESTORE \
  -DCCL \
  -DDESCQUORUM_OVERRIDE \
  -DDEVOPEN_REWRITE \
  -DDIO \
  -DDIO_XW \
  -DDISK_LEASE_DMS \
  -DDMAPI \
  -DDMAPI_MOVERS \
  -DDMAPI_PREFETCH_DIRATTRS \
  -DDMAPI_SYNC_DESTROY_EVENT \
  -DDMM_DISK \
  -DEE_WATCHDOG_START \
  -DENABLE_IMMUTABLE_FILES \
  -DEXPORTER_EVENTS \
  -DFASTEA \
  -DFILESETMAPPING_API \
  -DFINE_GRAINED_DIR_LOCKING \
  -DFSCK_INSTRUMENT \
  -DGPFS321_ADMIN \
  -DGPFS_CACHE \
  -DGPFS_QDIRTY \
  -DINODE64_PREP \
  -DLC_FENCING \
  $(DEFINE_LINUXPR) \
  -DMMAP_DIO \
  -DMMPMON \
  -DMMPMON_CUMULATIVE \
  -DMMPMON_DAEMON_PORT \
  -DMMPMON_HISTOGRAMS \
  -DMMPMON_NAMED_SOCKET \
  -DMMPMON_RPC \
  -DNEW_PFSCK \
  -DNFS4_ACL \
  -DNFS4_ACL_AUDIT \
  -DNFS4_VCM \
  -DNLINK32 \
  -DPARALLELQUOTACHK \
  -DPDEFRAG \
  $(DEFINE_PERCPU_BUFFER_TRACE_ONAIX) \
  -DPERF_STATS \
  -DPOLICY_RESTORE \
  -DPOW2_COND_SHIFT \
  $(DEFINE_PR_4_IPV6) \
  -DPTH_DISABLE_CANCELLATION \
  -DRESTORE_DMAPI \
  -DSANERGY \
  -DSFS_PERF_SKIP_SYNC \
  -DSFS_PERF_USE_SMALL_BUFFER \
  -DSIDMAPPING \
  -DSINGLE_PORT \
  -DSMB_LOCKS \
  -DTSC_SCATTERED_RECEIVE_REPLY_BUFFS \
  -DTSC_SCATTERED_SEND_REPLY_BUFFS \
  $(DEFINE_TSTRCFMT) \
  -DUIDREMAP \
  -DUSER_EXIT_ENTENSIONS

COMMON_TEMPSHIP_DEFINES = \
  -DDBGASSERTS \
  -DNEW_DELETE_DEBUG  \
  -DTRACK_USECOUNT \
  -DTRIGGERS \
  -DVERBOSETRACE

COMMON_NEVERSHIP_DEFINES = \
  -DENTRYEXIT_TRACE \
  -DINSTRUMENT_LOCKS \
  -DINSTRUMENT_TIME_LOCKS \
  -DMMDEBUG_MSGS \
  -DPTHREAD_STATS \
  -DTRACE_IO_DATA

#define DMAPI
#define EXPORTER_EVENTS
#define GPFS_CACHE
#define MMPMON
#define MMPMON_CUMULATIVE
#define MMPMON_HISTOGRAMS
#define PERF_STATS
#define SANERGY
/* #define SANERGY_RM */ /* needed when -DSANERGY is removed */
#define SMB_LOCKS
/* #define TRY_AND_BUY */
/* #define VDISK */

/*************************/
/* Solaris configuration */
/*************************/
#ifdef GPFS_SOLARIS

/* 64-bit (default) Solaris build.  All kernel files are compiled 64 bit */
/* All user-space is also built 64-bit, but it would be nice not to have to */
/* Debug build type default for now */
#define BLDTYPE DEBUG

#if BLDTYPE == DEBUG
  OPTFLAGS    = -g -mt -xarch=v9 -D__64BIT__ -xcode=abs64
  KOPTFLAGS   = -g -mt -xarch=v9 -DDEBUG -D__64BIT__
  CPLUSFLAGS  = -compat=5 -library=%none -xnolib -features=extensions
  KCPLUSFLAGS = -compat=5 -library=%none -xnolib -features=extensions -noex
  EXTRASTD_CFLAGS = -g -xarch=v9
  EXTRASTD_LDFLAGS = -g -xarch=v9
  EXTRALDFLAGS2 = 
  EXTRAARFLAGS = 
  EXTRAM4FLAGS = -D__64BIT__
  EXTRAASFLAGS = 
  MKTRCFMT_FLAGS = 
  EXTRAINSTROOTFLAGS =
  EXTRAINSTPROGFLAGS =
#else
  OPTFLAGS    = -xO2 -mt -xarch=v9 -D__64BIT__ -xcode=abs64
  KOPTFLAGS   = -xO2 -mt -xarch=v9 -D__64BIT__
  CPLUSFLAGS  = -compat=5 -library=%none -xnolib -features=extensions
  KCPLUSFLAGS = -compat=5 -library=%none -xnolib -features=extensions -noex
  EXTRASTD_CFLAGS = -xarch=v9
  EXTRASTD_LDFLAGS = -xarch=v9
  EXTRALDFLAGS2 = 
  EXTRAARFLAGS = 
  EXTRAM4FLAGS = -D__64BIT__
  EXTRAASFLAGS = 
  MKTRCFMT_FLAGS =
  EXTRAINSTROOTFLAGS = -s
  EXTRAINSTPROGFLAGS = -s
#endif

SITE_OTHER_KINCLUDES =
SITE_OTHER_KDEFINES = -D_KERNEL

SITE_FAR_TRACE_OPTS =
SITE_KFAR_TRACE_OPTS =

DEFAULT_DEFINES =  \
  -D_POSIX_PTHREAD_SEMANTICS \
  -D__EXTENSIONS__ \
  -D__FULL_PROTO   \
  -DGPFS_SOLARIS   \
  -DNO_ASM_SYNCH   \
  -DSSEG_SWIZZLE_PTRS \
  -DMUST_ALIGN     \
  -DSOLARIS_ACL_AS_SECATTR

TEMPSHIP_DEFINES = \
  
#define SSEG_SWIZZLE_PTRS

#endif /* end of SOLARIS */

/*********************/
/* AIX configuration */
/*********************/
#ifdef GPFS_AIX
#define AIX5

/* 64-bit AIX build (on HACMP) */
#define GPFSBLD64

#ifdef GPFSBLD64
  OPTFLAGS = -g
  KOPTFLAGS = -g
  EXTRASTD_CFLAGS = -q64 -qlargetype -qsuppress=1540-0053 -qsuppress=1540-5344 -qsuppress=1540-5345 -qsuppress=1540-2901 -qsuppress=1506-446
  EXTRASTD_LDFLAGS = -q64 -g
  EXTRASTD_LDFLAGS32 = -q32 -g
  EXTRASTD_LDFLAGS64 = -q64 -g
  EXTRALDFLAGS2 = -b64
  EXTRAARFLAGS = -X64
  EXTRAM4FLAGS = -D__64BIT__
  EXTRAASFLAGS = -a64
  MKTRCFMT_FLAGS = -q64
#else
  OPTFLAGS =
  KOPTFLAGS =
  EXTRASTD_CFLAGS = -qsuppress=1540-0053 -qsuppress=1540-5344 -qsuppress=1540-5345 -qsuppress=1540-2901 -qsuppress=1506-446 -qsuppress=1506-412
  EXTRASTD_LDFLAGS32 = -q32 -g
  EXTRASTD_LDFLAGS64 = -q64 -g
  MKTRCFMT_FLAGS = -q64
#endif

/* Build for AIX4 or AIX5
 * To build on AIX 5.x, you'll need these ifdefs
 *   -DSPAIXLVL=516 -DUSE51BREGS
 */
#ifndef AIX5
#define AIX4
#endif

#ifdef AIX4
  AIX_VERSION_DEPS = -DSPAIXLVL=432
#else
  AIX_VERSION_DEPS = -DSPAIXLVL=530 -DUSE51BREGS
#endif

SITE_OTHER_KINCLUDES =
SITE_OTHER_KDEFINES = -D_KERNEL

SITE_FAR_TRACE_OPTS =
SITE_KFAR_TRACE_OPTS =

DEFAULT_DEFINES = \
  $(AIX_VERSION_DEPS) \
  -D__FULL_PROTO \
  -D_LARGE_THREADS \
  -DGPFS_AIX

TEMPSHIP_DEFINES = \
  -DNEW_DELETE_TRACEBACKS

EXTRAINSTROOTFLAGS =
EXTRAINSTPROGFLAGS =

#endif /* end of AIX */

/***********************/
/* Linux configuration */
/***********************/
#ifdef GPFS_LINUX

SITE_OTHER_KDEFINES = -D_KERNEL -D__KERNEL__ -DMODULE \
                      -DGPFS_EXPORT_KSYMS -DEXPORT_SYMTAB

COMMON_LINUX = \
  -U_GNU_SOURCE \
  -D__USE_BSD \
  -D_LARGEFILE64_SOURCE \
  -DGPFS_LINUX \
  -DYESSTR=__YESSTR \
  -DNOSTR=__NOSTR \
  -DLUM \
  -DSSEG_SWIZZLE_PTRS \
  -D$(LINUX_DISTRIBUTION) \
  $(LINUX_PATCH_DEFINES)
#ifdef GPL_KBUILD
COMMON_LINUX += -DGPL_KBUILD
#endif
#ifdef LXTRACE_WANT_RELAY 
COMMON_LINUX += -DLXTRACE_WANT_RELAY
LXTRACE_WANT_RELAY_KBUILD = Y
#endif
#define LUM
#define SSEG_SWIZZLE_PTRS
#define MMANTRAS
#define MMANTRAS_CIM
/*#define CTDB*/

EXTRASTD_LDFLAGS = 
EXTRALDFLAGS2 =  

SITEMCRREV_RAW = "$Revision: 1.695.2.19 $ "

/* If you add -g to OPTFLAGS, you must also comment out the define of
 * FAR_TRACE.  It is recommended to remove the -fomit-frame-pointer 
 * option for better debugging information.  If you add -g to KOPTFLAGS, you 
 * must comment out the define of KFAR_TRACE. 
 */

/**********************/
/* Linux i386 defines */
/**********************/
#ifdef GPFS_ARCH_I386

OPTFLAGS  = -g0 -O2 -march=i686 -fno-strict-aliasing -fno-exceptions \
            -DGPFS_FRAME_POINTER

KOPTFLAGS = -O2 -march=i686 -fno-strict-aliasing -fno-exceptions \
            -fno-common -fomit-frame-pointer

#if LINUX_KERNEL_VERSION > 2060500

#if LINUX_KERNEL_VERSION >= 2061600
COMMON_26_FLAGS += -m32
KCONLYFLAGS += -ffreestanding
KOPTFLAGS += $(COMMON_26_FLAGS) -mpreferred-stack-boundary=2                                                               
OPTFLAGS  += $(COMMON_26_FLAGS)                                                                                            
#endif /* LINUX_KERNEL_VERSION >= 2061600 */
                                                                                  
#ifdef PCE_26_PCE
#define NOKREGPARMS
KOPTFLAGS += -mpreferred-stack-boundary=2 -DPCE_26_PCE
OPTFLAGS  += -DPCE_26_PCE            
#endif /* PCE_26_PCE */

#ifndef NOKREGPARMS
KOPTFLAGS += -mregparm=3 -mpreferred-stack-boundary=2
#else
SITE_OTHER_KDEFINES += -DNOKREGPARMS
#endif
#endif /* LINUX_KERNEL_VERSION > 2060500 */

#define KFAR_TRACE

DEFAULT_DEFINES = \
  $(COMMON_LINUX) \
  -DLINUX_KERNEL_VERSION=LINUX_KERNEL_VERSION \
  -DLIMIT_KSTACKS \
  -DSSEG_RESERVE_MEM

NEVERSHIP_DEFINES = -DKSTACK_CHECK -DSMART_PTR_DEBUG

#endif /* end of Linux i386 defines */


/***********************/
/* Linux Power defines */
/***********************/
#ifdef GPFS_ARCH_POWER
#define GPFS_BLUEGENE_P     /* for now all ppc is BG io */

OPTFLAGS = -g0 -O2 -fno-exceptions -fPIC
KOPTFLAGS = -O2 -msoft-float -fno-exceptions -fno-common

#undef KFAR_TRACE

DEFAULT_DEFINES = \
  $(COMMON_LINUX) \
  -DNO_ASM_SYNCH \
  -DSSEG_RESERVE_MEM \
  -DLINUX_KERNEL_VERSION=LINUX_KERNEL_VERSION

TEMPSHIP_DEFINES = -DDISABLE_KERNEL_PANIC
NEVERSHIP_DEFINES = -DKSTACK_CHECK -DSMART_PTR_DEBUG

#define NO_ASM_UTIL



#ifndef GPFS_BLUEGENE_P
  DEFAULT_DEFINES += -DCONFIG_UP
#endif
 
#ifdef GPFS_BLUEGENE_P
DEFAULT_DEFINES += -DGPFS_BLUEGENE_P
#define CROSS_COMPILE
#undef powerpc
CROSS_PREFIX = /bgsys/drivers/ppcfloor/gnu-linux/bin/powerpc-bgp-linux-
#endif /* end of GPFS_BLUEGENE_P cross compile */

#endif /* end of Linux power defines */


/**********************/
/* Linux ia64 defines */
/**********************/
#ifdef GPFS_ARCH_IA64

OPTFLAGS = -g0 -O2 -fno-strict-aliasing -fno-exceptions
KOPTFLAGS = -O2 -fno-strict-aliasing -fno-exceptions -fno-common

#undef KFAR_TRACE
#undef FAR_TRACE

DEFAULT_DEFINES = \
  $(COMMON_LINUX) \
  -DLINUX_KERNEL_VERSION=LINUX_KERNEL_VERSION 

NEVERSHIP_DEFINES = -DKSTACK_CHECK -DSMART_PTR_DEBUG

#endif /* end of Linux ia64 defines */

/***************************/
/* Linux PowerPC64 defines */
/***************************/
#ifdef GPFS_ARCH_PPC64

#if LINUX_KERNEL_VERSION < 2060000
CC = powerpc64-linux-g++
C_COMP = powerpc64-linux-gcc
CPP = powerpc64-linux-cpp
BARELD = powerpc64-linux-ld
AR = powerpc64-linux-ar clq
AS = powerpc64-linux-as
KAR = powerpc64-linux-ld $(LDFLAGS2) -r -o
RANLIB = powerpc64-linux-ranlib
ARDEL = $(AR) d
MAKEDEPEND = $(CC)
CPLUSINC = -I/opt/cross/include
#endif

OPTFLAGS = -g0 -O2 -mminimal-toc -fno-exceptions
KOPTFLAGS = -O -msoft-float -fno-exceptions -fno-common

#if LINUX_KERNEL_VERSION > 2060000
OPTFLAGS += -m64 -mpowerpc64
MAKEDEPEND = $(CC) -m64
#if (LINUX_KERNEL_VERSION >= 2061600)
KOPTFLAGS += -m64 -mcall-aixdesc -mno-altivec -mtune=power4 -funit-at-a-time -mstring -fno-ivopts -Wa,-maltivec
KCONLYFLAGS += -ffreestanding
#else
KOPTFLAGS += -m64 -mpowerpc64
#endif

EXTRASTD_LDFLAGS = -melf64ppc -m64
EXTRALDFLAGS2 =  -melf64ppc
BARELD += -melf64ppc

ifeq ("$(LINUX_DISTRIBUTION)","REDHAT_AS_LINUX")
#if LINUX_KERNEL_VERSION >= 2060905
KOPTFLAGS += -mcall-aixdesc
#endif
endif

#endif

DEFAULT_DEFINES = \
  $(COMMON_LINUX) \
  -DLINUX_KERNEL_VERSION=LINUX_KERNEL_VERSION

NEVERSHIP_DEFINES = -DKSTACK_CHECK -DSMART_PTR_DEBUG

#endif /* end of Linux PowerPC64 defines */

/**************************************/
/* Linux Opteron (aka x86_64) defines */
/**************************************/
#ifdef GPFS_ARCH_X86_64

OPTFLAGS = -g0 -O2 -fno-strict-aliasing -fno-exceptions -fPIC \
           -fomit-frame-pointer
KOPTFLAGS = -O2 -fno-strict-aliasing -fno-exceptions -fno-common \
            -mcmodel=kernel -mno-red-zone

CFLAGS32 = -m32 -U__64BIT__
SHLDFLAGS32 = -melf_i386

#undef KFAR_TRACE
#undef FAR_TRACE
#define API_32BIT

DEFAULT_DEFINES = \
  $(COMMON_LINUX) \
  -DLINUX_KERNEL_VERSION=LINUX_KERNEL_VERSION \
  -DLIMIT_KSTACKS

/* SMART_PTR_DEBUG turned off in ODE build due to large performance hit */
NEVERSHIP_DEFINES = -DKSTACK_CHECK -DSMART_PTR_DEBUG

#endif /* end of Linux x86_64 defines */

/***********************/
/* Linux s390x defines */
/***********************/
#ifdef GPFS_ARCH_S390X

OPTFLAGS  = -g0 -O2 -fomit-frame-pointer -fPIC \
	    -fno-strict-aliasing -fno-exceptions -mbackchain
KOPTFLAGS = -g0 -O2 -fomit-frame-pointer -fno-common \
	    -fno-strict-aliasing -fno-exceptions -mbackchain \
	    -mpacked-stack -msoft-float

#undef KFAR_TRACE
#undef FAR_TRACE

DEFAULT_DEFINES = \
  $(COMMON_LINUX) \
  -DLIMIT_KSTACKS \
  -DLINUX_KERNEL_VERSION=LINUX_KERNEL_VERSION

NEVERSHIP_DEFINES = -DKSTACK_CHECK -DSMART_PTR_DEBUG

#endif /* end of Linux s390x defines */

#ifdef FAR_TRACE
SITE_FAR_TRACE_OPTS = -DFAR_TRACE -DFAR_CODE -fno-defer-pop
#else
SITE_FAR_TRACE_OPTS =
#endif

#ifdef KFAR_TRACE
SITE_KFAR_TRACE_OPTS = -DFAR_TRACE -DFAR_CODE -fno-defer-pop
#else
SITE_KFAR_TRACE_OPTS =
#endif

#endif /* end of Linux defines */

/* Build complete list of -D compile flags */
SITE_DEFINES = $(ARCH_DEFINES) $(DEFAULT_DEFINES) $(COMMON_DEFINES)

/* kernel source build */
#ifdef SRC_BUILD
SITE_DEFINES += -DSRC_BUILD
#endif

/* libgpfs source build */
#ifdef LIBSRC_BUILD
SITE_DEFINES += -DLIBSRC_BUILD
#endif


/* Force compiler options needed for gprof if necessary */
#ifdef GPROF_DAEMON
CDEBUGFLAGS = -g -pg
#endif
